package routing

import (
	"encoding/json"
	"net/http"
	"strconv"
	"strings"

	"github.com/freakmaxi/kertish-dfs/basics/common"
	"go.uber.org/zap"
)

func (m *managerRouter) handlePut(w http.ResponseWriter, r *http.Request) {
	action := r.Header.Get("X-Action")

	if !m.validatePutAction(action) {
		w.WriteHeader(422)
		return
	}

	switch action {
	case "snapshot":
		m.handleRestoreSnapshot(w, r)
	default:
		w.WriteHeader(406)
	}
}

func (m *managerRouter) handleRestoreSnapshot(w http.ResponseWriter, r *http.Request) {
	clusterId, snapshotIndex, err := m.describeRestoreSnapshotOptions(r.Header.Get("X-Options"))
	if len(clusterId) == 0 || err != nil {
		w.WriteHeader(422)
		if err != nil {
			e := common.NewError(400, err.Error())
			if err := json.NewEncoder(w).Encode(e); err != nil {
				m.logger.Error("Response of restore snapshot request is failed", zap.Error(err))
			}
		}
	}

	err = m.manager.RestoreSnapshot(clusterId, snapshotIndex)
	if err == nil {
		return
	}

	w.WriteHeader(400)
	m.logger.Error("Restore snapshot request is failed", zap.Error(err))

	e := common.NewError(400, err.Error())
	if err := json.NewEncoder(w).Encode(e); err != nil {
		m.logger.Error("Response of restore snapshot request is failed", zap.Error(err))
	}
}

func (m *managerRouter) validatePutAction(action string) bool {
	switch action {
	case "snapshot":
		return true
	}
	return false
}

func (m *managerRouter) describeRestoreSnapshotOptions(options string) (string, uint64, error) {
	clusterId := ""
	eqIdx := strings.Index(options, "=")
	if eqIdx > -1 {
		clusterId = options[:eqIdx]
		options = options[eqIdx+1:]
	}
	snapshotIndex, err := strconv.ParseUint(options, 10, 64)
	if err != nil {
		return "", 0, err
	}

	return clusterId, snapshotIndex, nil
}
